@@ -10,17 +10,34 @@
 """
 Define type hints protocols for XPath related objects.
 """
-from typing import overload, Any, Dict, Iterator, Iterable, Optional, Sequence, ItemsView, \
-    Protocol, Sized, Hashable, Union, TypeVar, Mapping, Tuple, Set, MutableMapping
+from collections.abc import Hashable, Iterator, Iterable, ItemsView, Mapping, Sequence, Sized
+from typing import overload, Any, Optional, Protocol, Union, TypeVar, TYPE_CHECKING
+from xml.etree.ElementTree import Element, ElementTree
+
+from collections.abc import MutableMapping
+from elementpath.aliases import NamespacesType, NsmapType
+
+if TYPE_CHECKING:
+    from elementpath.schema_proxy import AbstractSchemaProxy
 
 _T = TypeVar("_T")
+_AnyStr = Union[str, bytes]
+
+
+class LxmlQNameProtocol(Protocol):
+    localname: _AnyStr
+    namespace: _AnyStr
+    text: _AnyStr
+
+
+LxmlKeyType = Union[str, bytes, LxmlQNameProtocol]
 
 
 class LxmlAttribProtocol(Protocol):
-    """A minimal protocol for attribute of lxml Element objects."""
+    """A minimal protocol for attributes of lxml Element objects."""
     def get(self, *args: Any, **kwargs: Any) -> Optional[str]: ...
 
-    def items(self) -> Sequence[Tuple[Any, Any]]: ...
+    def items(self) -> Sequence[tuple[Any, Any]]: ...
 
     def __contains__(self, key: Any) -> bool: ...
 
@@ -31,21 +48,13 @@ class LxmlAttribProtocol(Protocol):
     def __len__(self) -> int: ...
 
 
-AttribType = Union[
-    MutableMapping[str, Any],
-    MutableMapping[Optional[str], Any],
-    LxmlAttribProtocol,
-    'XsdAttributeGroupProtocol'
-]
-
-
 class ElementProtocol(Sized, Hashable, Protocol):
     """A protocol for generic ElementTree elements."""
 
     def __iter__(self) -> Iterator['ElementProtocol']: ...
 
     def find(
-            self, path: str, namespaces: Optional[Dict[str, str]] = ...
+            self, path: str, namespaces: Optional[dict[str, str]] = ...
     ) -> Optional['ElementProtocol']: ...
     def iter(self, tag: Optional[str] = ...) -> Iterator['ElementProtocol']: ...
 
@@ -55,6 +64,8 @@ class ElementProtocol(Sized, Hashable, Protocol):
     @overload
     def get(self, key: str, default: _T) -> Union[str, _T]: ...
 
+    def get(self, key: str, default: Optional[_T] = None) -> Union[str, _T, None]: ...
+
     @property
     def tag(self) -> str: ...
 
@@ -65,7 +76,7 @@ class ElementProtocol(Sized, Hashable, Protocol):
     def tail(self) -> Optional[str]: ...
 
     @property
-    def attrib(self) -> AttribType: ...
+    def attrib(self) -> 'AttribType': ...
 
 
 class EtreeElementProtocol(ElementProtocol, Protocol):
@@ -73,12 +84,12 @@ class EtreeElementProtocol(ElementProtocol, Protocol):
     def __iter__(self) -> Iterator['EtreeElementProtocol']: ...
 
     def find(
-            self, path: str, namespaces: Optional[Dict[str, str]] = ...
+            self, path: str, namespaces: Optional[dict[str, str]] = ...
     ) -> Optional['EtreeElementProtocol']: ...
     def iter(self, tag: Optional[str] = ...) -> Iterator['EtreeElementProtocol']: ...
 
     @property
-    def attrib(self) -> Dict[str, str]: ...
+    def attrib(self) -> dict[str, str]: ...
 
 
 class LxmlElementProtocol(ElementProtocol, Protocol):
@@ -86,7 +97,7 @@ class LxmlElementProtocol(ElementProtocol, Protocol):
     def __iter__(self) -> Iterator['LxmlElementProtocol']: ...
 
     def find(
-            self, path: str, namespaces: Optional[Dict[str, str]] = ...
+            self, path: str, namespaces: Optional[MutableMapping[str, str]] = ...
     ) -> Optional['LxmlElementProtocol']: ...
     def iter(self, tag: Optional[str] = ...) -> Iterator['LxmlElementProtocol']: ...
 
@@ -98,7 +109,7 @@ class LxmlElementProtocol(ElementProtocol, Protocol):
                      preceding: bool = False) -> Iterable['LxmlElementProtocol']: ...
 
     @property
-    def nsmap(self) -> Dict[Optional[str], str]: ...
+    def nsmap(self) -> NsmapType: ...
 
     @property
     def attrib(self) -> LxmlAttribProtocol: ...
@@ -168,6 +179,18 @@ class XsdTypeProtocol(XsdComponentProtocol, Protocol):
         """
         ...
 
+    def is_atomic(self) -> bool:
+        """Returns `True` if the instance is an atomic simpleType, `False` otherwise."""
+        ...
+
+    def is_list(self) -> bool:
+        """Returns `True` if the instance is a list simpleType, `False` otherwise."""
+        ...
+
+    def is_union(self) -> bool:
+        """Returns `True` if the instance is a union simpleType, `False` otherwise."""
+        ...
+
     def is_key(self) -> bool:
         """Returns `True` if it's a simpleType derived from xs:ID, `False` otherwise."""
         ...
@@ -180,6 +203,14 @@ class XsdTypeProtocol(XsdComponentProtocol, Protocol):
         """Returns `True` if it's a simpleType derived from xs:NOTATION, `False` otherwise."""
         ...
 
+    @overload
+    def is_valid(self, obj: Any, use_defaults: bool = True,
+                 namespaces: Optional[NamespacesType] = None,
+                 *args: Any, **kwargs: Any) -> bool: ...
+
+    @overload
+    def is_valid(self, obj: Any, *args: Any, **kwargs: Any) -> bool: ...
+
     def is_valid(self, obj: Any, *args: Any, **kwargs: Any) -> bool:
         """
         Validates an XML object node using the XSD type. The argument *obj* is an element
@@ -188,6 +219,14 @@ class XsdTypeProtocol(XsdComponentProtocol, Protocol):
         """
         ...
 
+    @overload
+    def validate(self, obj: Any, use_defaults: bool = True,
+                 namespaces: Optional[NamespacesType] = None,
+                 *args: Any, **kwargs: Any) -> None: ...
+
+    @overload
+    def validate(self, obj: Any, *args: Any, **kwargs: Any) -> None: ...
+
     def validate(self, obj: Any, *args: Any, **kwargs: Any) -> None:
         """
         Validates an XML object node using the XSD type. The argument *obj* is an element
@@ -213,6 +252,14 @@ class XsdTypeProtocol(XsdComponentProtocol, Protocol):
         """
         ...
 
+    @property
+    def simple_type(self) -> Optional['XsdTypeProtocol']:
+        """
+        The instance if it's a simpleType instance or the simpleType instance used for
+        deriving a complexType with simple content, `None` otherwise.
+        """
+        ...
+
 
 class XsdAttributeProtocol(XsdComponentProtocol, Protocol):
 
@@ -234,6 +281,9 @@ class XsdAttributeGroupProtocol(XsdComponentProtocol, Protocol):
     @overload
     def get(self, key: Optional[str], default: _T) -> Union[XsdAttributeProtocol, _T]: ...
 
+    def get(self, key: Optional[str], default: Optional[_T] = None) \
+        -> Union[XsdAttributeProtocol, _T, None]: ...
+
     def items(self) -> ItemsView[Optional[str], XsdAttributeProtocol]: ...
 
     def __contains__(self, key: Optional[str]) -> bool: ...
@@ -255,7 +305,7 @@ class XsdElementProtocol(XsdComponentProtocol, ElementProtocol, Protocol):
     def __iter__(self) -> Iterator['XsdElementProtocol']: ...
 
     def find(
-            self, path: str, namespaces: Optional[Dict[str, str]] = ...
+            self, path: str, namespaces: Optional[NamespacesType] = ...
     ) -> Optional[XsdXPathNodeType]: ...
     def iter(self, tag: Optional[str] = ...) -> Iterator['XsdElementProtocol']: ...
 
@@ -271,9 +321,12 @@ class XsdElementProtocol(XsdComponentProtocol, ElementProtocol, Protocol):
     @property
     def attrib(self) -> XsdAttributeGroupProtocol: ...
 
+    @property
+    def xpath_proxy(self) -> 'AbstractSchemaProxy': ...
+
 
 GT = TypeVar("GT")
-XsdGlobalValue = Union[GT, Tuple[ElementProtocol, Any]]
+XsdGlobalValue = Union[GT, tuple[ElementProtocol, Any]]
 
 
 class GlobalMapsProtocol(Protocol):
@@ -288,7 +341,7 @@ class GlobalMapsProtocol(Protocol):
     def elements(self) -> Mapping[str, XsdGlobalValue[XsdElementProtocol]]: ...
 
     @property
-    def substitution_groups(self) -> Mapping[str, Set[Any]]: ...
+    def substitution_groups(self) -> Mapping[str, set[Any]]: ...
 
 
 class XsdSchemaProtocol(XsdValidatorProtocol, ElementProtocol, Protocol):
@@ -296,19 +349,41 @@ class XsdSchemaProtocol(XsdValidatorProtocol, ElementProtocol, Protocol):
     def __iter__(self) -> Iterator[XsdXPathNodeType]: ...
 
     def find(
-            self, path: str, namespaces: Optional[Dict[str, str]] = ...
+            self, path: str, namespaces: Optional[NamespacesType] = ...
     ) -> Optional[XsdXPathNodeType]: ...
     def iter(self, tag: Optional[str] = ...) -> Iterator[XsdXPathNodeType]: ...
 
+    @property
+    def validity(self) -> str: ...
+
+    @property
+    def validation_attempted(self) -> str: ...
+
     @property
     def tag(self) -> str: ...
 
     @property
     def attrib(self) -> MutableMapping[Optional[str], 'XsdAttributeProtocol']: ...
 
+    @property
+    def xpath_proxy(self) -> 'AbstractSchemaProxy': ...
+
+
+DocumentType = Union[ElementTree, DocumentProtocol]
+ElementType = Union[Element, ElementProtocol]
+SchemaElemType = Union[XsdSchemaProtocol, XsdElementProtocol]
+CommentType = Union[Element, ElementProtocol]
+ProcessingInstructionType = Union[Element, ElementProtocol]
+AttribType = Union[
+    MutableMapping[str, Any],
+    MutableMapping[Optional[str], Any],
+    LxmlAttribProtocol,
+    XsdAttributeGroupProtocol
+]
 
 __all__ = ['ElementProtocol', 'EtreeElementProtocol', 'LxmlAttribProtocol',
            'LxmlElementProtocol', 'DocumentProtocol', 'LxmlDocumentProtocol',
            'XsdValidatorProtocol', 'XsdComponentProtocol', 'XsdTypeProtocol',
-           'XsdAttributeProtocol', 'XsdAttributeGroupProtocol',
-           'XsdElementProtocol', 'GlobalMapsProtocol', 'XsdSchemaProtocol',]
+           'XsdAttributeProtocol', 'XsdAttributeGroupProtocol', 'XsdElementProtocol',
+           'GlobalMapsProtocol', 'XsdSchemaProtocol', 'DocumentType', 'ElementType',
+           'SchemaElemType', 'CommentType', 'ProcessingInstructionType', 'AttribType']
