@@ -8,33 +8,35 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 from abc import ABCMeta, abstractmethod
-from typing import TYPE_CHECKING, cast, Any, Dict, List, Optional, Iterator, Union
+from collections.abc import Iterator
+from functools import lru_cache
+from typing import TYPE_CHECKING, Any, Optional, Union
 
-from .exceptions import ElementPathTypeError
-from .protocols import ElementProtocol, XsdTypeProtocol, XsdAttributeProtocol, \
+from elementpath.exceptions import ElementPathTypeError
+from elementpath.protocols import XsdTypeProtocol, XsdAttributeProtocol, \
     XsdElementProtocol, XsdSchemaProtocol
-from .datatypes import AtomicValueType
-from .etree import is_etree_element
-from .xpath_context import XPathSchemaContext
+from elementpath.datatypes import AtomicType
+from elementpath.etree import is_etree_element
+from elementpath.xpath_context import XPathSchemaContext
 
 if TYPE_CHECKING:
-    from .xpath2 import XPath2Parser
-    from .xpath30 import XPath30Parser
+    from elementpath.xpath_tokens import XPath2ParserType
 
-    XPathParserType = Union[XPath2Parser, XPath30Parser]
-else:
-    XPathParserType = Any
+PathResult = Union[XsdSchemaProtocol, XsdElementProtocol, XsdAttributeProtocol]
 
 
 class AbstractSchemaProxy(metaclass=ABCMeta):
     """
-    Abstract base class for defining schema proxies.
+    Abstract base class for defining schema proxies. An implementation can override
+    initialization type annotations
 
-    :param schema: a schema instance that implements the `AbstractEtreeElement` interface.
+    :param schema: a schema instance compatible with the XsdSchemaProtocol.
     :param base_element: the schema element used as base item for static analysis.
     """
+    __slots__ = ('_schema', '_base_element', '_find', '_is_fully_valid')
+
     def __init__(self, schema: XsdSchemaProtocol,
-                 base_element: Optional[ElementProtocol] = None) -> None:
+                 base_element: Optional[XsdElementProtocol] = None) -> None:
         if not is_etree_element(schema):
             raise ElementPathTypeError(
                 "argument {!r} is not a compatible schema instance".format(schema)
@@ -45,9 +47,44 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
             )
 
         self._schema = schema
-        self._base_element: Optional[ElementProtocol] = base_element
+        self._base_element: Optional[XsdElementProtocol] = base_element
 
-    def bind_parser(self, parser: XPathParserType) -> None:
+        if self._base_element is not None:
+            self._find = self._base_element.find
+        else:
+            self._find = self._schema.find
+
+        self._is_fully_valid = False
+
+    @property
+    def schema(self) -> XsdSchemaProtocol:
+        return self._schema
+
+    @property
+    def base_element(self) -> Optional[XsdElementProtocol]:
+        return self._base_element
+
+    @property
+    def validity(self) -> str:
+        validity = self._schema.validity
+        if validity != 'valid':
+            self._is_fully_valid = False
+        return validity
+
+    @property
+    def validation_attempted(self) -> str:
+        validation_attempted = self._schema.validation_attempted
+        if validation_attempted != 'full':
+            self._is_fully_valid = False
+        return validation_attempted
+
+    def is_fully_valid(self) -> bool:
+        if self._is_fully_valid:
+            return True
+        self._is_fully_valid = self.validity == 'valid' and self.validation_attempted == 'full'
+        return self._is_fully_valid
+
+    def bind_parser(self, parser: 'XPath2ParserType') -> None:
         """
         Binds a parser instance with schema proxy adding the schema's atomic types constructors.
         This method can be redefined in a concrete proxy to optimize schema bindings.
@@ -67,10 +104,10 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
 
         :returns: an `XPathSchemaContext` instance.
         """
-        return XPathSchemaContext(root=self._schema, item=self._base_element)
+        return XPathSchemaContext(root=self._schema, item=self._base_element, schema=self)
 
-    def find(self, path: str, namespaces: Optional[Dict[str, str]] = None) \
-            -> Optional[XsdElementProtocol]:
+    def find(self, path: str, namespaces: Optional[dict[str, str]] = None) \
+            -> Optional[PathResult]:
         """
         Find a schema element or attribute using an XPath expression.
 
@@ -78,13 +115,28 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
         :param namespaces: an optional mapping from namespace prefix to namespace URI.
         :return: The first matching schema component, or ``None`` if there is no match.
         """
-        return cast(Optional[XsdElementProtocol], self._schema.find(path, namespaces))
+        return self._find(path, namespaces)
+
+    @lru_cache(maxsize=None)
+    def cached_find(self, expanded_path: str) -> Optional[PathResult]:
+        """
+        Find a schema element or attribute using an expanded path as XPath expression.
+
+        :param expanded_path: an XPath expression with qualified names already resolved \
+        to expanded form.
+        :return: The first matching schema component, or ``None`` if there is no match.
+        """
+        return self._find(expanded_path)
 
     @property
     def xsd_version(self) -> str:
         """The XSD version, returns '1.0' or '1.1'."""
         return self._schema.xsd_version
 
+    def is_assertion_based(self) -> bool:
+        return self._base_element is not None and \
+            self._base_element.parent is self._base_element.type
+
     def get_type(self, qname: str) -> Optional[XsdTypeProtocol]:
         """
         Get the XSD global type from the schema's scope. A concrete implementation must
@@ -94,7 +146,10 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
         :param qname: the fully qualified name of the type to retrieve.
         :returns: an object that represents an XSD type or `None`.
         """
-        return self._schema.maps.types.get(qname)
+        xsd_type = self._schema.maps.types.get(qname)
+        if isinstance(xsd_type, tuple):
+            return None
+        return xsd_type
 
     def get_attribute(self, qname: str) -> Optional[XsdAttributeProtocol]:
         """
@@ -105,7 +160,10 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
         :param qname: the fully qualified name of the attribute to retrieve.
         :returns: an object that represents an XSD attribute or `None`.
         """
-        return self._schema.maps.attributes.get(qname)
+        xsd_attribute = self._schema.maps.attributes.get(qname)
+        if isinstance(xsd_attribute, tuple):
+            return None
+        return xsd_attribute
 
     def get_element(self, qname: str) -> Optional[XsdElementProtocol]:
         """
@@ -116,13 +174,17 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
         :param qname: the fully qualified name of the element to retrieve.
         :returns: an object that represents an XSD element or `None`.
         """
-        return self._schema.maps.elements.get(qname)
+        xsd_element = self._schema.maps.elements.get(qname)
+        if isinstance(xsd_element, tuple):
+            return None
+        return xsd_element
 
-    def get_substitution_group(self, qname: str) -> Optional[List[XsdElementProtocol]]:
+    def get_substitution_group(self, qname: str) -> Optional[set[XsdElementProtocol]]:
         """
-        Get a substitution group. A concrete implementation must returns a list containing
-        substitution elements or `None` if the substitution group is not found. Moreover each item
-        of the returned list must be an object that implements the `AbstractXsdElement` interface.
+        Get a substitution group. A concrete implementation must return a list containing
+        substitution elements or `None` if the substitution group is not found. Moreover,
+        each item of the returned list must be an object that implements the
+        `AbstractXsdElement` interface.
 
         :param qname: the fully qualified name of the substitution group to retrieve.
         :returns: a list containing substitution elements or `None`.
@@ -139,7 +201,7 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
         """
 
     @abstractmethod
-    def cast_as(self, obj: Any, type_qname: str) -> AtomicValueType:
+    def cast_as(self, obj: Any, type_qname: str) -> AtomicType:
         """
         Converts *obj* to the Python type associated with an XSD global type. A concrete
         implementation must raises a `ValueError` or `TypeError` in case of a decoding
@@ -157,4 +219,4 @@ class AbstractSchemaProxy(metaclass=ABCMeta):
         """
 
 
-__all__ = ['AbstractSchemaProxy']
+__all__ = ['PathResult', 'AbstractSchemaProxy']
